// Copyright 2016 Orus.io. All rights reserved.

package auth

import (
	"encoding/pem"
	"fmt"
	"runtime/debug"
	"strings"

	"github.com/nats-io/gnatsd/server"
	"github.com/nats-io/go-nats"
	"github.com/nats-io/nuid"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewXBusAuth returns a gnatsd.server.Auth implementation on
// top of actor & account storage instances.
// It provides a simple login/password authentication based on the Account model stored in the
// database.
// A special token allow full-access (used by internal services)
func NewXBusAuth(
	accountStorage storage.AccountStorage, actorStorage storage.ActorStorage, logger xbus.Logger,
) XBusAuth {
	return XBusAuth{
		AccountStorage: accountStorage,
		ActorStorage:   actorStorage,
		AuthToken:      nuid.Next(),
		Logger:         logger,
	}
}

// XBusAuth ...
type XBusAuth struct {
	AccountStorage storage.AccountStorage
	ActorStorage   storage.ActorStorage
	AuthToken      string
	Logger         xbus.Logger
}

func defaultPermissions(token string) *server.Permissions {
	return &server.Permissions{
		Publish: []string{"xbus.registration.register"},
		Subscribe: []string{
			nats.InboxPrefix + "anon-" + token + ".in.*",
		},
	}
}

// Check authenticate a client for first or third auth step
func (a *XBusAuth) Check(c server.ClientAuthentication) bool {
	// We do not want storage errors to crash the server
	defer func() {
		if r := recover(); r != nil {
			a.Logger.Errorf("Error during identity Check: %s\n%s", r, debug.Stack())
		}
	}()

	tlsState := c.GetTLSConnectionState()

	if tlsState == nil {
		// non-TLS connections are no longer allowed
		a.Logger.Errorf("Non-TLS connection")
		return false
	}

	opts := c.GetOpts()

	if a.AuthToken != "" && opts.Authorization == a.AuthToken {
		return true
	}

	if len(tlsState.VerifiedChains) == 0 {
		a.Logger.Noticef("No client certificate -> anonymous connection")
		user := server.User{
			Username:    "anonymous",
			Permissions: defaultPermissions(opts.Username),
		}
		c.RegisterUser(&user)
		return true
	}

	var account api.Account
	var ok bool

	for _, chain := range tlsState.VerifiedChains {
		cert := chain[0]
		pCert := pem.EncodeToMemory(&pem.Block{
			Type:  "CERTIFICATE",
			Bytes: cert.Raw,
		})
		account, ok = a.AccountStorage.GetByCert(string(pCert))
		if ok {
			a.Logger.Debugf("Found account '%s' matching certificate", account.Name)
			break
		}
	}

	if !ok {
		a.Logger.Errorf("Unknown certificate.")
		return false
	}

	var apiKey string

	if account.Type == api.Account_GATEWAY {
		if strings.HasPrefix(opts.Authorization, "-----BEGIN CERTIFICATE-----") {
			account, ok = a.AccountStorage.GetByCert(opts.Authorization)
		} else if opts.Authorization != "" {
			apiKey = opts.Authorization
			account, ok = a.AccountStorage.GetByAPIKey(opts.Authorization)
		}
		if !ok {
			a.Logger.Errorf("auth_token matches no account: %s", opts.Authorization)
			return false
		}
	}

	// Default permissions any account has.
	permissions := server.Permissions{
		Publish: []string{
			nats.InboxPrefix + account.ID.String() + ".out.*",
			nats.InboxPrefix + account.ID.String() + ".in.*.heartbeat",
			api.ClientPrefix + account.ID.String() + ".*",
			"xbus.registration.update." + account.ID.String(),
		},
		Subscribe: []string{
			nats.InboxPrefix + account.ID.String() + ".in.*",
		},
	}

	if apiKey != "" {
		whoamiSubject, whoamiReply := api.WhoAmISubjects(apiKey)
		permissions.Publish = append(
			permissions.Publish,
			whoamiSubject)
		permissions.Subscribe = append(
			permissions.Subscribe,
			whoamiReply,
		)
	}

	var err error
	switch account.Type {
	case api.Account_ACTOR:
		err = fillActorPermissions(account, &permissions, a)
	case api.Account_USER:
		err = fillUserPermissions(account, &permissions)
	default:
		a.Logger.Errorf("Invalid account type: %s\n", account.Type)
		return false
	}

	if err != nil {
		a.Logger.Errorf("Error filling permissions: %s", err)
		return false
	}

	a.Logger.Debugf("%s permissions: %+v", account.Name, permissions)
	user := server.User{Username: account.Name, Permissions: &permissions}
	c.RegisterUser(&user)
	a.Logger.Noticef("Authenticated %s", account.Name)

	return true
}

// Fill permissions Xbus actors have access to.
func fillActorPermissions(
	account api.Account, permissions *server.Permissions, a *XBusAuth) error {

	// Got an account; find corresponding actors to set permissions accordingly.
	actors := a.ActorStorage.ListAccount(account.ID)

	// Add specific permissions for each actor.
	for _, actor := range actors {
		switch actor.Kind {
		case api.Actor_EMITTER:
			permissions.Publish = append(
				permissions.Publish,
				api.ActorOutputSubject(actor.ID, "*"),
			)
			permissions.Subscribe = append(
				permissions.Subscribe,
				api.EmitterEnvelopeStateSubject(actor.ID, api.UUID{}),
			)
		case api.Actor_CONSUMER:
			permissions.Subscribe = append(
				permissions.Subscribe,
				api.ActorInputSubject(actor.ID, "*"),
				api.ActorInputSubject(actor.ID, "*")+".*")
		case api.Actor_WORKER:
			permissions.Subscribe = append(
				permissions.Subscribe,
				api.ActorInputSubject(actor.ID, "*"),
				api.ActorInputSubject(actor.ID, "*")+".*")
			permissions.Publish = append(
				permissions.Publish,
				api.ActorOutputSubject(actor.ID, "*"))
		default:
			return fmt.Errorf("Invalid actor type: %s", actor.Kind)
		}
		permissions.Publish = append(
			permissions.Publish,
			"xbus.sentinel.actorready."+actor.ID.String(),
			"xbus.actor."+actor.ID.String()+".>",
		)
	}
	return nil
}

// Fill permissions regular API users have access to.
func fillUserPermissions(account api.Account, permissions *server.Permissions) error {
	permissions.Publish = append(permissions.Publish, "xbus.control.>")
	return nil
}
