// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"
	"time"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/utils"
)

// accountCmd represents the account command.
var accountCmd = &cobra.Command{
	Use:   "account",
	Short: "Manage accounts that are allowed to connect to Xbus.",
	Long: `This command contains several sub-commands to manage accounts and the
permissions they have within Xbus.

In particular, it is possible from this interface to accept / reject pending
account requests.`,
}

// accountListCmd represents the account list sub-command.
var accountListCmd = &cobra.Command{
	Use:   "list",
	Short: "List accounts.",
	Long:  "Display accounts allowed to connect to Xbus, as a list.",
	Run:   listAccounts,
}

// accountAcceptCmd represents the account accept sub-command.
var accountAcceptCmd = &cobra.Command{
	Use:   "accept <identifier>",
	Short: "Accept an account.",
	Long:  "Accept a pending account request so it can be allowed to connect to Xbus.",
	Run:   acceptAccount,
}

// accountRejectCmd represents the account reject sub-command.
var accountRejectCmd = &cobra.Command{
	Use:   "reject <identifier>",
	Short: "Reject an account.",
	Long:  "Reject a pending account request; it will not be allowed to connect to Xbus.",
	Run:   rejectAccount,
}

// accountDeleteCmd represents the account delete sub-command.
var accountDeleteCmd = &cobra.Command{
	Use:   "delete <identifier>",
	Short: "Delete an account.",
	Long:  "Delete an account.",
	Run:   deleteAccount,
}

var accountRenewAPIKeyCmd = &cobra.Command{
	Use:   "renew-apikey",
	Short: "Renew the APIKey of an account",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 1 {
			cmd.Usage()
			os.Exit(1)
		}
		os.Exit(accountRenewAPIKey(args[0]))
	},
}

// Arguments of the "accept" command.
const (
	acceptCmdArgFingerprint = iota
	acceptCmdArgCount
)

// Arguments of the "reject" command.
const (
	rejectCmdArgFingerprint = iota
	rejectCmdArgCount
)

// Arguments of the "delete" command.
const (
	deleteCmdArgIdentifier = iota
	deleteCmdArgCount
)

// listAccounts parses arguments to list accounts.
func listAccounts(cmd *cobra.Command, args []string) {
	csrHash := viper.GetBool("csr-hash")
	certDetails := viper.GetBool("cert-details")
	os.Exit(doListAccounts(csrHash, certDetails))
}

// doListAccounts lists accounts.
func doListAccounts(csrHash bool, certDetails bool) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	accounts, err := conn.AccountList()
	if err != nil {
		Logger.Errorf("Cannot retrieve accounts: %s", err)
		return 1
	}

	accountCount := len(accounts)
	if accountCount == 0 {
		fmt.Println("No account found.")
	} else {
		fmt.Printf("%d accounts found:\n", accountCount)
		for _, account := range accounts {
			fmt.Println(fmtAccount(account, csrHash, certDetails))
		}
	}
	return 0
}

// acceptAccount parses arguments to accept a pending account request.
func acceptAccount(cmd *cobra.Command, args []string) {
	if len(args) < acceptCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", acceptCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	expire, err := getCertExpire(cmd)
	if err != nil {
		Logger.Errorf("%s", err)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doAcceptAccount(args[acceptCmdArgFingerprint], expire))
}

// doAcceptAccount accepts a pending account request.
func doAcceptAccount(identifier string, expire time.Time) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	id, err := findAccountID(conn, identifier)
	if err == xbus.ErrNotFound {
		Logger.Errorf("No such account: %s", identifier)
		return 1
	} else if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}

	// Got a match! Send an "accept" request.
	repAccount, err := conn.AccountAccept(id, expire)
	if err != nil {
		Logger.Errorf("Could not accept the account %q: %s", id, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Errorf("Could not accept the account %q: Still pending.", id)
		return 1
	}
	Logger.Noticef("Successfully accepted the account %q.", id)
	return 0
}

// rejectAccount parses arguments to reject a pending account request.
func rejectAccount(cmd *cobra.Command, args []string) {
	if len(args) < rejectCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", rejectCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doRejectAccount(args[rejectCmdArgFingerprint]))
}

// deleteAccount parses arguments to delete an account
func deleteAccount(cmd *cobra.Command, args []string) {
	if len(args) < deleteCmdArgCount {
		Logger.Errorf("Not enough arguments; %d arguments are required.", deleteCmdArgCount)
		cmd.Usage()
		os.Exit(1)
	}

	os.Exit(doDeleteAccount(args[deleteCmdArgIdentifier]))
}

// doRejectAccount rejects a pending account request.
func doRejectAccount(identifier string) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	id, err := findAccountID(conn, identifier)
	if err == xbus.ErrNotFound {
		Logger.Errorf("No such account: %s", identifier)
		return 1
	} else if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}

	// Got a match! Send a "reject" request.
	repAccount, err := conn.AccountReject(id)
	if err != nil {
		Logger.Errorf("Could not reject the account %q: %s", id, err)
		return 1
	}
	if repAccount.Status == api.Account_PENDING {
		Logger.Errorf("Could not reject the account %q: Still pending.", id)
		return 1
	}
	Logger.Noticef("Successfully rejected the account %q.", id)
	return 0
}

// doDeleteAccount deletes an account
func doDeleteAccount(identifier string) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not delete account %q: %s", identifier, err)
		return 1
	}

	// Got a match! Send a "delete" request.
	err = conn.AccountDelete(account.ID)
	if err != nil {
		Logger.Errorf("Could not delete the account %q: %s", account.Name, err)
		return 1
	}
	return 0
}

func accountRenewAPIKey(identifier string) int {
	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	account, err := conn.AccountFind(identifier)
	if err != nil {
		Logger.Errorf("Could not find account %s: %s", identifier, err)
		return 1
	}

	account, err = conn.AccountRenewAPIKey(account.ID)
	if err != nil {
		Logger.Errorf("Error renewing the account: %s", err)
		return 1
	}

	Logger.Noticef("New API Key:\n%s", account.APIKey)

	return 0
}

func findAccountID(conn *xbus.Client, identifier string) (api.UUID, error) {
	if id, err := api.UUIDFromString(identifier); err == nil {
		return id, nil
	}
	account, err := conn.AccountFind(identifier)
	return account.ID, err
}

func fmtAccount(account api.Account, csrHash bool, certDetails bool) string {
	formattedAccount := fmt.Sprintf(`
ID: %s
Name: %s
Type: %s
Status: %s`,
		account.ID,
		account.Name,
		account.Type,
		account.Status,
	)
	if csrHash {
		formattedAccount += fmt.Sprintf(`
CSR fingerprint: %s`,
			tools.GetCSRHash(account.CSR, true),
		)
	}
	if certDetails {
		formattedAccount += fmt.Sprintf(`
CSR: %s
Cert: %s`,
			account.CSR,
			account.Cert,
		)
	}
	return formattedAccount
}

func getCertExpire(cmd *cobra.Command) (time.Time, error) {
	expireString, err := cmd.Flags().GetString("cert-expire")
	if err != nil {
		return time.Time{}, err
	}
	return utils.ParseDateOrDuration(expireString)
}

func init() {
	RootCmd.AddCommand(accountCmd)

	accountCmd.AddCommand(accountListCmd)
	accountListFlags := accountListCmd.PersistentFlags()
	accountListFlags.Bool("cert-details", false, "Display cert details")
	viper.BindPFlag("cert-details", accountListFlags.Lookup("cert-details"))
	accountListFlags.Bool("csr-hash", true, "Display CSR fingerprints")
	viper.BindPFlag("csr-hash", accountListFlags.Lookup("csr-hash"))

	flags := accountAcceptCmd.Flags()
	flags.String("cert-expire", "1y", "Expire date or duration for the account certificate")

	viper.BindPFlag("accountaccept.cert-expire", flags.Lookup("cert-expire"))

	accountCmd.AddCommand(accountAcceptCmd)
	accountCmd.AddCommand(accountRejectCmd)
	accountCmd.AddCommand(accountDeleteCmd)
	accountCmd.AddCommand(accountRenewAPIKeyCmd)
}
