// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"io"
	"io/ioutil"
	"os"

	"bitbucket.org/orus-io/go-xbus/api"

	"github.com/spf13/cobra"
)

func graphInfo(identifier, version string) api.GraphInfo {
	var info api.GraphInfo
	if id, err := api.UUIDFromString(identifier); err == nil {
		info.ID = id
	} else {
		info.Name = identifier
		info.Version = version
	}
	return info
}

func graphQuery(cmd *cobra.Command, args []string) {
	c, err := GetClient()
	if err != nil {
		return
	}
	if len(args) > 1 {
		cmd.Usage()
		os.Exit(-1)
	}
	var name string
	if len(args) == 1 {
		name = args[0]
	}

	graphs, err := c.GraphQuery(name)
	if err != nil {
		Logger.Errorf("%s", err)
		return
	}
	for _, info := range graphs {
		fmt.Printf("%v\n", info)
	}
}

func graphGet(identifier, version, filename string) error {
	c, err := GetClient()
	if err != nil {
		return err
	}

	info := graphInfo(identifier, version)

	str, err := c.GraphGet(info)
	if err != nil {
		return err
	}

	var out io.Writer
	if filename == "-" {
		out = os.Stdout
	} else {
		f, err := os.Create(filename)
		if err != nil {
			return err
		}
		defer f.Close()
		out = f
	}
	out.Write([]byte(str))
	return nil
}

func graphSave(identifier, version, filename string) error {
	c, err := GetClient()
	if err != nil {
		return err
	}

	content, err := ioutil.ReadFile(filename)
	if err != nil {
		return fmt.Errorf("Cannot read graph file: %s", err)
	}

	info := graphInfo(identifier, version)

	info, graph, validationMessages, err := c.GraphSave(info, string(content))
	if len(validationMessages) != 0 {
		fmt.Println("Validation messages:")
		for _, msg := range validationMessages {
			fmt.Println("-", msg)
		}
	}

	if err != nil {
		return err
	}

	if err := ioutil.WriteFile(filename, []byte(graph), 0660); err != nil {
		return fmt.Errorf("Cannot save the resulting graph: %s", err)
	}

	Logger.Noticef("Successfully saved graph %s-%s", info.Name, info.Version)
	return nil
}

func graphStatus(identifier, version string, status api.GraphInfo_Status) error {
	c, err := GetClient()
	if err != nil {
		return err
	}

	info := graphInfo(identifier, version)
	info.Status = status

	info, err = c.GraphSetStatus(info)
	if err != nil {
		return err
	}
	Logger.Noticef("%s-%s status changed to %s", info.Name, info.Version, info.Status)
	return nil
}

var graphCmd = &cobra.Command{
	Use:   "graph",
	Short: "Graph control subcommands",
	Long:  "Graph subcommands allow to list and modify graphs",
}

var graphQueryCmd = &cobra.Command{
	Use:   "query <name>",
	Short: "Search for graphs with the given name",
	Run:   graphQuery,
}

var graphGetCmd = &cobra.Command{
	Use:   "get <name or id> [version]",
	Short: "Get a graph definition",
	Long:  "Get a graph definition and write it to stdout",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) == 0 || len(args) > 2 {
			cmd.Usage()
			os.Exit(-1)
		}
		var (
			identifier = args[0]
			version    string
			filename   = cmd.Flag("output").Value.String()
		)
		if len(args) == 2 {
			version = args[1]
		}
		if err := graphGet(identifier, version, filename); err != nil {
			Logger.Errorf("Error getting graph: %s", err)
			os.Exit(-1)
		}
	},
}

var graphSaveCmd = &cobra.Command{
	Use:   "save <name or id> [version] <filepath>",
	Short: "Save a graph",
	Long:  "Save a graph definition",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) < 2 || len(args) > 3 {
			cmd.Usage()
			os.Exit(-1)
		}
		var (
			identifier = args[0]
			version    string
			filename   string
		)
		if len(args) == 2 {
			filename = args[1]
		} else {
			version = args[1]
			filename = args[2]
		}
		if err := graphSave(identifier, version, filename); err != nil {
			Logger.Errorf("Error saving graph: %s", err)
			os.Exit(-1)
		}
	},
}

var graphStatusCmd = &cobra.Command{
	Use:   "status <name or id> [version] [--active|--inactive]",
	Short: "Change the status of a graph version",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) < 1 || len(args) > 2 {
			cmd.Usage()
			os.Exit(-1)
		}
		var (
			identifier = args[0]
			version    string
		)
		if len(args) == 2 {
			version = args[1]
		}
		var status api.GraphInfo_Status
		if cmd.Flag("active").Value.String() == "true" {
			status = api.GraphInfo_ACTIVE
		} else if cmd.Flag("inactive").Value.String() == "true" {
			status = api.GraphInfo_INACTIVE
		}
		if err := graphStatus(identifier, version, status); err != nil {
			Logger.Errorf("Error setting graph status: %s", err)
			os.Exit(-1)
		}
	},
}

func init() {
	RootCmd.AddCommand(graphCmd)
	graphCmd.AddCommand(graphQueryCmd)
	graphCmd.AddCommand(graphGetCmd)
	graphCmd.AddCommand(graphSaveCmd)
	graphCmd.AddCommand(graphStatusCmd)

	graphStatusCmd.Flags().Bool("active", false, "Activate the graph")
	graphStatusCmd.Flags().Bool("inactive", false, "Deactivate the graph")

	graphGetCmd.Flags().StringP("output", "o", "-", "Target filename (default to stdout)")
}
