// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"io/ioutil"
	"os"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

// registerCmd represents the "register" command.
var registerCmd = &cobra.Command{
	Use:   "register",
	Short: "Generate a CSR and ask Xbus for a registration.",
	Long: `Generate a Certificate Signing Request then send it over to Xbus through the
unencrypted "registration" channel. This xbusctl will only be allowed to send
Xbus actual commands once accepted on the Xbus side.
This command also outputs the command one has to run on the Xbus side to accept
the registration request.`,
	Run: register,
}

// register parses arguments to proceed with the registration.
func register(cmd *cobra.Command, args []string) {
	certDir := viper.GetString("cert-dir")
	login := viper.GetString("user.login")
	natsURL := viper.GetString("nats.url")
	os.Exit(doRegister(login, natsURL, certDir))
}

// loadPrivKey loads the private key stored in the specified file.
func loadPrivKey(privKeyPath string) (privKey interface{}, err error) {
	privKeyData, err := ioutil.ReadFile(privKeyPath)
	if err != nil {
		return nil, err
	}
	privKeyBlock, _ := pem.Decode(privKeyData)
	return x509.ParseECPrivateKey(privKeyBlock.Bytes)
}

// doRegister generates a CSR then asks Xbus for a registration.
func doRegister(login string, natsURL string, certDir string) int {
	xbusClient, err := GetClient()
	if err != nil {
		Logger.Errorf("%s", err)
		return 1
	}

	// TODO find the actual hostname or get one in the configuration
	// TODO Read org and cn from the configuration
	if err := xbusClient.CreateCSR("localhost", "xbus", "xbusctl", false); err != nil {
		if err != xbus.ErrAlreadyExists {
			Logger.Errorf("Cannot generate a CSR: %s", err)
			return 1
		}
	}

	// Send the registration request.
	regStatus, err := xbusClient.Register()
	if err != nil {
		Logger.Errorf("Cannot send the registration request: %s", err)
		return 1
	}

	Logger.Noticef("Successfully sent the registration request.")

	// Say something depending on the registration status we have received in reply.
	switch regStatus {
	case api.Registration_PENDING:
		Logger.Noticef(`The registration is pending approval. To approve from Xbus, run:
	xbusd account accept %q`,
			//tools.GetCSRHash(string(csrData), false),
			"xxxx",
		)
	case api.Registration_ACCEPTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically accepted.",
		)
	case api.Registration_REJECTED:
		Logger.Noticef(
			"The registration was already known by Xbus; it has been automatically rejected.",
		)
	}

	return 0
}

// saveAccountID saves the specified account ID into a file.
func saveAccountID(accountID string, login string, certDir string) error {
	return ioutil.WriteFile(fmt.Sprintf("%s/%s.id", certDir, login), []byte(accountID), 0644)
}

// saveCert saves the specified certificate (provided as raw DER data) to a file.
func saveCert(certData []byte, login string, certDir string) error {
	certFile, err := os.Create(fmt.Sprintf("%s/%s.crt", certDir, login))
	if err != nil {
		return err
	}
	err = pem.Encode(certFile, &pem.Block{Type: "CERTIFICATE", Bytes: certData})
	if err != nil {
		return err
	}
	err = certFile.Close()
	if err != nil {
		return err
	}
	return nil
}

func init() {
	RootCmd.AddCommand(registerCmd)
}
