package fullenv

import (
	"regexp"
	"testing"
	"time"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"github.com/stretchr/testify/assert"
)

func newBasRelayActor(actor *xbus.Actor) xbus.ActorService {
	return &badRelayActor{actor, nil}
}

type badRelayActor struct {
	actor *xbus.Actor
	sub   *xbus.Subscription
}

func (a *badRelayActor) Startup() error {
	sub, err := a.actor.Subscribe(func(er *xbus.EnvelopeReceiver) {
		env, err := er.Complete(time.Second)
		if err != nil {
			a.actor.GetClient().Log.Errorf("Reception error: %s", err)
			return
		}
		status, err := a.actor.SendEnvelope("default", api.ForwardEnvelope(env))
		if status == api.EnvelopeAck_ACCEPTED && err == nil {
			a.actor.GetClient().Log.Noticef("Successfully forwarded the envelope")
		} else {
			a.actor.GetClient().Log.Errorf("Something wrong happened. Status=%s, err=%s", status, err)
		}
		if err := a.actor.ProcessingEnd(
			env.Context, env.ID, api.ActorProcessingEndRequest_SUCCESS, nil,
		); err != nil {
			a.actor.GetClient().Log.Errorf("Error sending ProcessingEnd : %s", err)
		}
	})
	if err != nil {
		return err
	}
	a.sub = sub
	return nil
}

func (a *badRelayActor) Shutdown() error {
	err := a.sub.Unsubscribe()
	a.sub = nil
	return err
}

func TestInvalidProcessState(t *testing.T) {
	fe, _ := NewFullEnvForTest(t, "invalid-process-state", "invalid_process_state.yaml")

	if err := fe.Init(); err != nil {
		t.Fatal(err)
	}
	defer fe.Close()

	fe.RegisterActorService("badrelayactor", newBasRelayActor)

	fe.Startup()

	// Emit one message
	sender := fe.Emitter.Actor.OpenOutput(api.EnvelopeContext{}, "default", "msg")
	if err := sender.AddItems("msg", api.Item("hi")); err != nil {
		t.Fatal(err)
	}
	if err := sender.Close(); err != nil {
		t.Fatal(err)
	}

	// Wait for the process to be over
	time.Sleep(100 * time.Millisecond)

	re, err := regexp.Compile("\\| ([0-9a-f-]+) \\| [0-9a-f-]+ \\| [0-9a-f-]+ \\| [A-Z]+ +\\|") // EXTRACT PIDS)

	// The relay will misbehave, and the process should pass in a 'error' state
	matchLines, err := fe.XbusCtl.RunReadAndFindSubmatch(re, "ps", "--status-error")
	must(err)
	assert.Equal(t, 1, len(matchLines), matchLines)

	// Restarting xbus should not be an issue
}
