// Copyright 2016 Orus.io. All rights reserved.

package fullstack

import (
	"strconv"
	"sync"
	"sync/atomic"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
)

func TestSlowConsumer(t *testing.T) {
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	ft.InstallGraph(`
name: forward
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> target
`)

	var (
		itemCount     int
		fragmentCount int
	)

	csub, err := consumer.RawSubscribe(
		func(input string, envelope api.Envelope) (api.EnvelopeAck, error) {
			time.Sleep(50 * time.Millisecond)
			ack := api.EnvelopeAck{
				ID:     envelope.ID,
				Status: api.EnvelopeAck_ACCEPTED,
			}
			fragmentCount++
			itemCount += len(envelope.Events[0].Items)
			t.Log("Consumer received envelope -", itemCount, "items so far")
			if itemCount == 3000 {
				go consumer.ProcessingSuccess(envelope.Context, envelope.ID)
			}
			return ack, nil
		}, true)
	if err != nil {
		t.Fatal(err)
	}
	defer csub.Unsubscribe()

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	var item = make([]byte, 1024)

	go func() {
		output := emitter.OpenOutput(api.EnvelopeContext{}, "default", "evt1")
		for i := 0; i != 3000; i++ {
			output.AddItems("evt1", api.Item(item))
		}
		assert.Nil(t, output.Close())
		t.Log("Sent all the items")
	}()

	timeout := time.After(20 * time.Second)
	var processID api.UUID

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_RUNNING, state.Status)
		processID = state.ProcessList[0].ID
	case <-timeout:
		t.Errorf("Process never started")
		t.FailNow()
	}

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_DONE, state.Status)
		assert.Empty(t, state.ProcessList[0].Errors)
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	case <-timeout:
		t.Errorf("Process never ended. Current state:")
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	}
}

func TestStreamedEnvelope(t *testing.T) {
	const emissionLoops = 20000
	ft := InitFullstack(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})

	_, worker := ft.CreateWorker("fw", []string{"forward"})
	wsub, err := worker.Subscribe(func(r *xbus.EnvelopeReceiver) {
		worker.Log.Noticef("Receiving envelope")
		go func() {
			t.Log("context ----", r.Context())
			output := worker.OpenOutput(r.Context(), "default", "evt1", "evt2")
			c1 := r.GetItemChan("evt1")
			c2 := r.GetItemChan("evt2")
			wg := sync.WaitGroup{}
			wg.Add(2)
			go func() {
				for item := range c1 {
					if err := output.AddItems("evt1", item); err != nil {
						t.Error(err)
						break
					}
				}
				wg.Done()
			}()
			go func() {
				for item := range c2 {
					if err := output.AddItems("evt2", item); err != nil {
						t.Error(err)
						break
					}
				}
				wg.Done()
			}()
			wg.Wait()
			worker.Log.Noticef("Done receiving")
			assert.Nil(t, output.Close())
		}()
	})
	defer wsub.Unsubscribe()

	var (
		receivedEvt1Count int64
		receivedEvt2Count int64
	)

	csub, err := consumer.Subscribe(
		func(r *xbus.EnvelopeReceiver) {
			go func() {
				c1 := r.GetItemChan("evt1")
				c2 := r.GetItemChan("evt2")
				for receivedEvt1Count != emissionLoops && receivedEvt2Count != 3*emissionLoops {
					select {
					case <-c1:
						atomic.AddInt64(&receivedEvt1Count, 1)
					case <-c2:
						atomic.AddInt64(&receivedEvt2Count, 1)
					}
				}
				consumer.ProcessingSuccess(r.Context(), r.ID)
			}()
		})
	if err != nil {
		t.Fatal(err)
	}
	defer csub.Unsubscribe()

	ft.InstallGraph(`
name: forward
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1,evt2]
  outputs: [default]
- id: relay
  type: worker
  actors: [fw]
  inputs: [default]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> relay
- relay.default -> target.default
`)

	pStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer sub.Unsubscribe()

	go func() {
		output := emitter.OpenOutput(api.EnvelopeContext{}, "default", "evt1", "evt2")
		for i := 0; i != emissionLoops; i++ {
			output.AddItems("evt1", api.Item("evt1-item-"+strconv.Itoa(i)))
			output.AddItems("evt2",
				api.Item("evt2-0-item-"+strconv.Itoa(i)),
				api.Item("evt2-1-item-"+strconv.Itoa(i)),
				api.Item("evt2-2-item-"+strconv.Itoa(i)))
		}
		assert.Nil(t, output.Close())
	}()

	timeout := time.After(180 * time.Second)
	var processID api.UUID

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_RUNNING, state.Status)
		processID = state.ProcessList[0].ID
	case <-timeout:
		t.Errorf("Process never started")
		t.FailNow()
	}

	select {
	case state := <-pStateChan:
		assert.Equal(t, api.Process_DONE, state.Status)
		assert.Empty(t, state.ProcessList[0].Errors)
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	case <-timeout:
		t.Errorf("Process never ended. Current state:")
		pState := ft.Server.ProcessStorage().GetState(processID)
		t.Log(pState.Fmt())
	}
}
