package process

import (
	"fmt"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewManager creates a process.Manager
func NewManager(
	graphManager *graph.Manager,
	actorStorage storage.ActorStorage,
	processStorage storage.ProcessStorage,
	logger xbus.Logger,
) *Manager {
	manager := Manager{
		GraphManager:   graphManager,
		ActorStorage:   actorStorage,
		ProcessStorage: processStorage,
		Log:            xbus.NewPrefixedLogger("[process/manager] ", logger),
	}
	return &manager
}

// Manager manage the processes
type Manager struct {
	GraphManager   *graph.Manager
	ActorStorage   storage.ActorStorage
	ProcessStorage storage.ProcessStorage

	Log xbus.Logger

	CurrentProcessList []*Process
}

// Close does nothing for now
func (m *Manager) Close() {
}

func (m *Manager) loadProcess(p storage.Process) (*Process, error) {
	var g *graph.Graph
	if !p.GraphID.IsZero() {
		g1, exists := m.GraphManager.GetGraph(p.GraphID)
		if !exists {
			return nil, fmt.Errorf("No such graph: %s", p.GraphID)
		}
		g = &g1
	}
	return &Process{
		ID:                p.ID,
		actors:            m.ActorStorage,
		processStorage:    m.ProcessStorage,
		graphManager:      m.GraphManager,
		manager:           m,
		TriggerEmitterID:  p.TriggerEmitterID,
		TriggerEnvelopeID: p.TriggerEnvelopeID,
		Graph:             g,
		SourceRef:         p.SourceOutput,
	}, nil
}

// LookupOrCreate returns the Process handling the given envelope or create
// if necessary
func (m *Manager) LookupOrCreate(emitter api.Actor, envelopeID api.UUID) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcessByTrigger(emitter, envelopeID)
	if ok {
		return m.loadProcess(p)
	}

	p = storage.Process{
		TriggerEmitterID:  emitter.ID,
		TriggerEnvelopeID: envelopeID,
	}
	var err error
	p.ID, err = m.ProcessStorage.CreateProcess(p)
	if err != nil {
		return nil, err
	}
	return m.loadProcess(p)
}

// GetEmitterEnvelopeStatus returns the statuses of the processes triggered
// by the given emitter/envelope
func (m *Manager) GetEmitterEnvelopeStatus(emitter api.Actor, envelopeID api.UUID) map[api.UUID]api.Process_Status {
	p, ok := m.ProcessStorage.GetProcessByTrigger(emitter, envelopeID)
	if !ok {
		return nil
	}
	return map[api.UUID]api.Process_Status{
		p.ID: m.ProcessStorage.GetState(p.ID).Status,
	}
}

// Get returns a process from its ID
func (m *Manager) Get(id api.UUID) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcess(id)
	if !ok {
		return nil, fmt.Errorf("No such process: %s", id)
	}
	return m.loadProcess(p)
}

// ListActiveEmissions lists active emissions for a given target
func (m *Manager) ListActiveEmissions(target api.UUID) []storage.Emission {
	return m.ProcessStorage.ListActiveEmissions(target)
}

// ListProcessEmissions lists active emissions for a given process
func (m *Manager) ListProcessEmissions(process api.UUID) []storage.Emission {
	return m.ProcessStorage.ListProcessEmissions(process)
}

// ListPostMortem returns dead processes
func (m *Manager) ListPostMortem(level api.LogLevel, status ...api.PMProcess_Status) ([]*Process, error) {
	var result []*Process

	for _, sProcess := range m.ProcessStorage.QueryPostMortem(level, status...) {
		process, err := m.loadProcess(sProcess)
		if err != nil {
			return nil, err
		}
		result = append(result, process)
	}

	return result, nil
}

// Purge removes any trace of the process in the process storage.
func (m *Manager) Purge(id api.UUID) error {
	// TODO drop the logs
	return m.ProcessStorage.PurgeProcess(id)
}
