// Copyright 2017 Orus.io. All rights reserved.

package broker

import (
	"context"
	"fmt"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/storage"
)

func (s *Service) receiveEnvelope(actorID api.UUID, output string, envelope api.Envelope) (process *process.Process, sourceRef storage.NodeOutputRef, ack api.EnvelopeAck, err error) {
	s.Log.Debugf("Receiving envelope fragment (%s)", envelope.ID)

	ack.ID = envelope.ID

	actor, exists := s.Opts.ActorStorage.Get(actorID)
	if !exists {
		err = api.ClientProtoError("Unknown actor: %s", actorID)
		return
	}

	// If no context is given, make sure the sender is as emitter
	if envelope.Context.ProcessID.IsZero() && actor.Kind != api.Actor_EMITTER {
		msg := fmt.Sprintf("Context-less envelope sent by non-emitter %s", actor.Name)
		s.Log.Errorf(msg)
		err = api.ClientProtoError(msg)
		return
	}

	// A context is given, make sure the process exists
	if !envelope.Context.ProcessID.IsZero() {
		process, err = s.processManager.Get(envelope.Context.ProcessID)
		if err != nil {
			ack = *api.NewEnvelopeAckError(envelope.ID, err)
			err = nil
			return
		}
		// TODO make sure NodeID matches the sending actor
		sourceRef.NodeID = envelope.Context.NodeID
		sourceRef.Output = output
	}

	var (
		typesKnown bool
		eventTypes []string
	)

	// Check for empty envelope
	if envelope.Last == true && len(envelope.Events) == 0 && len(envelope.EventIDs) == 0 {
		if process == nil {
			ack = *api.NewEnvelopeAckError(
				envelope.ID,
				fmt.Errorf("Invalid envelop (missing or invalid context)"))
			return
		}

		// empty envelope, which can mean 2 things depending on the ID

		// Without an ID: the envelope is just a signal that the output
		// will not be used in this context. Inform the process.
		if envelope.ID.IsZero() {
			ack.Status = api.EnvelopeAck_ACCEPTED

			process.UpdateReceiveStatus(
				envelope.ID, actor.ID, sourceRef, storage.ReceptionClosed)
			return
		}
		// With an ID: the envelope is one sent on an input, hence must be
		// known to the broker which can reread it from storage.
		ack.Status = api.EnvelopeAck_ACCEPTED
		typesKnown = true
	} else {
		// non-empty envelope. Store it.
		var status storage.EnvelopeStatus
		status, err = s.Opts.EnvelopeStorage.StoreEnvelope(envelope)
		if err != nil {
			panic(err)
		}
		s.Log.Debugf("Stored envelope: %s", status.Status)
		switch status.Status {
		case "receiving":
			ack.Status = api.EnvelopeAck_RECEIVING
		case "complete":
			ack.Status = api.EnvelopeAck_ACCEPTED
		case "error":
			ack.Status = api.EnvelopeAck_ERROR
		default:
			ack.Status = api.EnvelopeAck_ERROR
		}
		typesKnown = status.TypesKnown
		eventTypes = status.EventTypes()

		if ack.Status == api.EnvelopeAck_ERROR {
			if process != nil {
				process.UpdateReceiveStatus(
					envelope.ID, actor.ID, sourceRef, storage.ReceptionError)
			}
			ack = *api.NewEnvelopeAckError(envelope.ID, status)
			return
		}
	}

	// The envelope is now safely stored
	// We just need the process to be aware of this reception, and we will
	// be done.

	if envelope.Context.ProcessID.IsZero() {
		process, err = s.processManager.LookupOrCreate(actor, envelope.ID)
		if err != nil {
			panic(fmt.Errorf("Error getting or creating process: %s", err))
		}
		if process.Graph != nil {
			sourceRef = process.SourceRef
		}
	}

	// No context but all types known -> the process can match a graph if needed
	if process.Graph == nil && envelope.Context.ProcessID.IsZero() && typesKnown {
		if matchErr := process.MatchGraph(eventTypes); matchErr != nil {
			ack.Status = api.EnvelopeAck_ERROR
			ack.Reason = fmt.Sprintf(matchErr.Error())
			s.Log.Errorf(ack.Reason)
			return
		}
		sourceRef = process.SourceRef
	}

	// If process is unknown and the reception is over, we have a serious
	// problem
	if process == nil && ack.Status == api.EnvelopeAck_ACCEPTED {
		panic("Impossible state, please report")
	}

	if process != nil && process.Graph != nil {
		// If a process is associated to the envelope, make sure the route for the next
		// move is known, then update the reception status.
		// Logically we should calculate the route after updating the status, but if the
		// server crashes we would have route-less receptions that would be hard to
		// track and fix. Switching the two operations is harmless: a reception-less
		// route will become used as soon as the emitter re-attempt its sending.
		// And because the route is present, the emission service will find what
		// to do as soon as it is (re)started
		// See issue #39

		_, err := process.GetTargets(sourceRef, envelope.ID, false)
		if err != nil {
			panic(fmt.Sprintf("Error getting routes: %s", err))
		}

		var storageReceptionStatus storage.ReceptionStatus
		switch ack.Status {
		case api.EnvelopeAck_RECEIVING:
			storageReceptionStatus = storage.ReceptionOngoing
		case api.EnvelopeAck_ACCEPTED:
			storageReceptionStatus = storage.ReceptionDone
		}
		if err := process.UpdateReceiveStatus(
			envelope.ID, actor.ID, sourceRef, storageReceptionStatus,
		); err != nil {
			panic(fmt.Sprintf("Error updating receive status: %s", err))
		}
	}

	// our part (receive the envelope fragment) is done, we can safely
	// let the ack be sent to the sender

	return
}

func (s *Service) routeEnvelope(
	process *process.Process,
	envelopeID api.UUID,
	sourceRef storage.NodeOutputRef,
) {
	targets, err := process.GetTargets(sourceRef, envelopeID, false)
	if err != nil {
		s.Log.Errorf("Error getting targets: %s", err)
		return
	}

	s.Log.Debugf("Will forward to %s", targets)

	c := api.NewEmissionClient(s.Conn)
	for _, target := range targets {
		c.Signal(api.EmissionSignal{
			ProcessID:  process.ID,
			EnvelopeID: envelopeID,
			Target: api.EnvelopeTarget{
				ActorID: target.ActorID,
				NodeID:  target.NodeID,
				Input:   target.Input,
			},
		})
	}
}

type brokerServer struct {
	*Service
}

// Output receives envelopes sent by actors to their outputs
func (s brokerServer) Output(
	ctx context.Context,
	output string, envelope api.Envelope,
) (
	api.EnvelopeAck, error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	actorSID := ctx.Value("nrpc-svc-actorID").(string)
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return api.EnvelopeAck{}, err
	}

	// XXX we do not want to block other incoming envelopes, we should have workers
	// doing the job in the background so we can handle a number of concurrent incoming
	// envelopes
	// We do not want to fire a route for each incoming envelope either,
	// to avoid handling too many envelopes at the same time

	// handle the reception
	process, sourceRef, ack, err := s.receiveEnvelope(actorID, output, envelope)

	// to the routing if relevant
	if err == nil && ack.Status != api.EnvelopeAck_ERROR &&
		!envelope.ID.IsZero() &&
		process != nil && process.Graph != nil {
		go s.routeEnvelope(process, envelope.ID, sourceRef)
	}

	return ack, err
}
