// Copyright 2017 Orus.io. All rights reserved.

package clientapi

import (
	"context"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

// QueueName is the queue name used by the service
const QueueName = "xbus.services.clientapi"

// Options for the Service
type Options struct {
	services.Options

	ActorStorage   storage.ActorStorage
	AccountStorage storage.AccountStorage
}

// NewService instanciate a Service
func NewService(o Options) *Service {
	return &Service{
		o:           o,
		BaseService: services.NewBaseService("clientapi", o.Options),
	}
}

// Service provides an api for administrators
type Service struct {
	services.BaseService

	o Options

	clientAPIHandler *api.ClientAPIHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	whoamiSubject, _ := api.WhoAmISubjects("*")
	s.clientAPIHandler = api.NewClientAPIHandler(
		context.Background(), s.Conn, clientAPIServer{s},
	)
	whoamiHandler := api.NewWhoAmIHandler(
		context.Background(), s.Conn, clientAPIServer{s},
	)
	// The two handlers have the same subject prefix 'xbus.client',
	// so we need to subscribe to their methods directly
	// TODO the api need some rework to avoid this situation
	if err := s.SubscribeAll(
		services.QueueSubscribe(
			"xbus.client.*.getactors", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			"xbus.client.*.getactors.*", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			whoamiSubject, QueueName,
			whoamiHandler.Handler),
	); err != nil {
		s.Shutdown()
		return err
	}
	return nil
}

type clientAPIServer struct {
	*Service
}

func (s clientAPIServer) GetActors(ctx context.Context) (
	resp api.ClientGetActorsReply, err error,
) {
	accountSID := ctx.Value("nrpc-svc-accountID").(string)
	accountID, err := api.UUIDFromString(accountSID)
	if err != nil {
		return
	}

	resp.Actors = s.o.ActorStorage.ListAccount(accountID)
	return
}

func (s clientAPIServer) WhoAmI(
	ctx context.Context, apiKey string,
) (api.Account, error) {
	account, ok := s.o.AccountStorage.GetByAPIKey(apiKey)
	if !ok {
		panic("No account by this api key")
	}
	return account, nil
}
