// Copyright 2016 Orus.io. All rights reserved.

package control

import (
	"context"
	"encoding/json"
	"fmt"
	"sync"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/api/control"
	"gopkg.in/yaml.v2"
)

type processServer struct {
	*Service
}

func (s processServer) Query(
	ctx context.Context, req control.ProcessQueryRequest,
	send func(control.ProcessQueryReply),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	pList := s.o.ProcessStorage.Query(req.Filter)
	select {
	case <-ctx.Done():
		return nil
	default:
	}
	var batch []api.Process
	for _, p := range pList {
		state := s.o.ProcessStorage.GetState(p.ID)
		b, err := json.Marshal(state)
		if err != nil {
			panic(err)
		}
		batch = append(batch, api.Process{
			ID:                p.ID,
			TriggerEmitterID:  p.TriggerEmitterID,
			TriggerEnvelopeID: p.TriggerEnvelopeID,
			GraphID:           p.GraphID,
			Status:            state.Status,
			State:             string(b),
		})
		select {
		case <-ctx.Done():
			return nil
		default:
		}
		if len(batch) == 1000 {
			send(control.ProcessQueryReply{ProcessList: batch})
			batch = nil
		}
	}
	if batch != nil {
		send(control.ProcessQueryReply{ProcessList: batch})
	}
	return nil
}

func (s processServer) Control(
	ctx context.Context, req control.ProcessControlRequest,
) (err error) {
	s.TaskAdd()
	defer s.TaskDone()
	switch req.Command {
	case control.ProcessControlRequest_PAUSE:
		err = s.o.ProcessStorage.SetStatus(req.ProcessID, api.Process_PAUSED)
	case control.ProcessControlRequest_RESUME:
		err = s.o.ProcessStorage.SetStatus(req.ProcessID, api.Process_RUNNING)
	default:
		err = fmt.Errorf("Invalid command: %s", err)
	}
	return
}

func (s processServer) Export(
	ctx context.Context, req control.ProcessExportRequest,
) (
	resp control.ProcessExportReply, err error,
) {
	s.TaskAdd()
	defer s.TaskDone()

	for _, pid := range req.ProcessIDs {
		p, err := s.o.ProcessManager.Get(pid)
		if err != nil {
			panic(err)
		}

		export, err := p.Export(s.o.EnvelopeStorage, s.Log)
		if err != nil {
			panic(err)
		}

		b, err := yaml.Marshal(export)
		if err != nil {
			panic(err)
		}

		resp.ProcessList = append(resp.ProcessList, string(b))
	}
	return
}

func (s processServer) Purge(
	ctx context.Context, req control.ProcessPurgeRequest,
	send func(control.TaskProgress),
) error {
	s.TaskAdd()
	defer s.TaskDone()

	if len(req.ProcessIDs) == 0 {
		send(control.TaskProgress{})
		return nil
	}

	var wg sync.WaitGroup
	jobCount := len(req.ProcessIDs) / 100
	if jobCount == 0 {
		jobCount = 1
	} else if jobCount > s.o.MaxPurgeJobs {
		jobCount = s.o.MaxPurgeJobs
	}

	jobsCh := make(chan api.UUID)
	resCh := make(chan error, jobCount)
	stopCh := make(chan struct{})

	wg.Add(jobCount + 1)

	for i := 0; i != jobCount; i++ {
		go func() {
			defer wg.Done()
			for {
				select {
				case pid, ok := <-jobsCh:
					if !ok {
						return
					}
					resCh <- s.o.ProcessManager.Purge(pid)
				case <-ctx.Done():
					return
				}
			}
		}()
	}

	var (
		max     = uint32(len(req.ProcessIDs))
		counter uint32
	)

	go func() {
		defer wg.Done()
		defer close(jobsCh)
		for _, pid := range req.ProcessIDs {
			select {
			case <-stopCh:
				return
			case <-ctx.Done():
				return
			case jobsCh <- pid:
			}
		}
	}()

	for {
		select {
		case <-ctx.Done():
			wg.Wait()
			return nil

		case err := <-resCh:
			if err != nil {
				stopCh <- struct{}{}
				wg.Wait()
				return err
			}
			counter++
			send(control.TaskProgress{Progression: counter, MaxProgression: max})
			if counter == max {
				wg.Wait()
				return nil
			}
		}
	}
}
