// Copyright 2016 Orus.io. All rights reserved.

package control

import (
	"context"
	"crypto/x509"

	"bitbucket.org/orus-io/go-xbus/api/control"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

var queueName = "xbus.control"

// Options for the Service
type Options struct {
	services.Options

	ServerCert *x509.Certificate
	ServerKey  interface{}

	MaxPurgeJobs int

	GraphManager    *graph.Manager
	ProcessManager  *process.Manager
	ProcessStorage  storage.ProcessStorage
	AccountStorage  storage.AccountStorage
	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	LogStorage      storage.LogStorage
}

// NewService instanciate a Service
func NewService(o Options) *Service {
	if o.MaxPurgeJobs == 0 {
		o.MaxPurgeJobs = 1
	}
	return &Service{
		o:           o,
		BaseService: services.NewBaseService("control", o.Options),
	}
}

// Service provides an api for administrators
type Service struct {
	services.BaseService

	accountHandler   *control.AccountHandler
	actorHandler     *control.ActorHandler
	envelopeHandler  *control.EnvelopeHandler
	graphHandler     *control.GraphHandler
	logsHandler      *control.LogsHandler
	pmprocessHandler *control.PMProcessHandler
	processHandler   *control.ProcessHandler
	o                Options
}

// Startup starts the service
func (s *Service) Startup() error {
	s.accountHandler = control.NewAccountHandler(
		context.Background(), s.Conn, accountServer{s},
	)
	s.actorHandler = control.NewActorHandler(
		context.Background(), s.Conn, actorServer{s},
	)
	s.envelopeHandler = control.NewEnvelopeHandler(
		context.Background(), s.Conn, envelopeServer{s},
	)
	s.graphHandler = control.NewGraphHandler(
		context.Background(), s.Conn, graphServer{s},
	)
	s.logsHandler = control.NewLogsHandler(
		context.Background(), s.Conn, logsServer{s},
	)
	s.pmprocessHandler = control.NewPMProcessHandler(
		context.Background(), s.Conn, pmprocessServer{s},
	)
	s.processHandler = control.NewProcessHandler(
		context.Background(), s.Conn, processServer{s},
	)
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(s.accountHandler, queueName),
		s.HandlerQueueSubscribe(s.actorHandler, queueName),
		s.HandlerQueueSubscribe(s.envelopeHandler, queueName),
		s.HandlerQueueSubscribe(s.graphHandler, queueName),
		s.HandlerQueueSubscribe(s.logsHandler, queueName),
		s.HandlerQueueSubscribe(s.processHandler, queueName),
		s.HandlerQueueSubscribe(s.pmprocessHandler, queueName),
	); err != nil {
		s.Shutdown()
		return err
	}
	return nil
}
