// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"encoding/json"
	"fmt"
	"strconv"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// NewProcessStorage initialize a new ProcessStorage
func NewProcessStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *ProcessStorage {
	return &ProcessStorage{
		Base: Base{
			model:  model,
			db:     db,
			logger: logger,
		},
		onStatusChanged:        storage.NoOpProcessEventStatusHandler,
		processLogEntryHandler: storage.VoidProcessLogEntryHandler,
	}
}

// ProcessStorage implements storage.ProcessStorage
type ProcessStorage struct {
	Base
	onStatusChanged        storage.ProcessEventStatusHandler
	processLogEntryHandler storage.ProcessLogEntryHandler
}

// SetEventHandlers sets the event handlers
func (s *ProcessStorage) SetEventHandlers(onStatusChanged storage.ProcessEventStatusHandler) {
	if onStatusChanged == nil {
		s.onStatusChanged = storage.NoOpProcessEventStatusHandler
	} else {
		s.onStatusChanged = onStatusChanged
	}
}

// SetProcessLogEntryHandler sets the process logentry handler
func (s *ProcessStorage) SetProcessLogEntryHandler(
	handler storage.ProcessLogEntryHandler,
) {
	if handler == nil {
		s.processLogEntryHandler = storage.VoidProcessLogEntryHandler
	} else {
		s.processLogEntryHandler = handler
	}
}

func dbProcessToStorage(dbProcess model.Process) storage.Process {
	return storage.Process{
		ID:                api.UUID(dbProcess.ID),
		TriggerEmitterID:  api.UUID(dbProcess.TriggerEmitterID),
		TriggerEnvelopeID: api.UUID(dbProcess.TriggerEnvelopeID),
		GraphID:           api.UUID(dbProcess.GraphID),
		SourceOutput: storage.NodeOutputRef{
			NodeID: dbProcess.SourceNodeID,
			Output: dbProcess.SourceSlotID,
		},
	}
}

func storageProcessToDB(process storage.Process) model.Process {
	return model.Process{
		Base:              model.Base{ID: uuid.UUID(process.ID)},
		TriggerEmitterID:  uuid.UUID(process.TriggerEmitterID),
		TriggerEnvelopeID: uuid.UUID(process.TriggerEnvelopeID),
		GraphID:           uuid.UUID(process.GraphID),
		SourceNodeID:      process.SourceOutput.NodeID,
		SourceSlotID:      process.SourceOutput.Output,
	}
}

func (s *ProcessStorage) loadProcess(tx *yago.Tx, id api.UUID) (storage.Process, bool) {
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Get(&dbProcess, uuid.UUID(id)); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

// GetProcess returns a process data
func (s *ProcessStorage) GetProcess(id api.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	return s.loadProcess(tx.Tx, id)
}

// GetProcessByTrigger returns a process matching a trigger or an empty value if not found
// TODO in the future we should support multiple process per trigger.
func (s *ProcessStorage) GetProcessByTrigger(emitter api.Actor, envelopeID api.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Where(
		s.model.Process.TriggerEmitterID.Eq(uuid.UUID(emitter.ID)),
		s.model.Process.TriggerEnvelopeID.Eq(uuid.UUID(envelopeID)),
	).One(&dbProcess); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

func uuidSliceToInterfaceSlice(inlist []api.UUID) (outlist []interface{}) {
	for _, v := range inlist {
		outlist = append(outlist, uuid.UUID(v))
	}
	return
}

func processStatusSliceToInterfaceSlice(inlist []api.Process_Status) (outlist []interface{}) {
	for _, v := range inlist {
		outlist = append(outlist, v)
	}
	return
}

// Query lookup for processes
func (s *ProcessStorage) Query(filter api.ProcessFilter) []storage.Process {
	var dbProcessList []model.Process
	tx := s.MustBegin()
	defer tx.Rollback()
	q := tx.Query(s.model.Process)
	if len(filter.ID) != 0 {
		q = q.Filter(s.model.Process.ID.In(
			uuidSliceToInterfaceSlice(filter.ID)...))
	}
	if len(filter.TriggerEmitterID) != 0 {
		q = q.Filter(s.model.Process.TriggerEmitterID.In(
			uuidSliceToInterfaceSlice(filter.TriggerEmitterID)...))
	}
	if len(filter.TriggerEnvelopeID) != 0 {
		q = q.Filter(s.model.Process.TriggerEnvelopeID.In(
			uuidSliceToInterfaceSlice(filter.TriggerEnvelopeID)...))
	}
	if len(filter.GraphID) != 0 {
		q = q.Filter(s.model.Process.GraphID.In(
			uuidSliceToInterfaceSlice(filter.GraphID)...))
	}

	if len(filter.Status) != 0 ||
		(filter.StatusChangedFrom != nil) ||
		(filter.StatusChangedTo != nil) {

		q = q.LeftJoin(s.model.ProcessState)

		if len(filter.Status) != 0 {
			q = q.Filter(
				s.model.ProcessState.Status.In(
					processStatusSliceToInterfaceSlice(filter.Status)...),
			)
		}

		if filter.StatusChangedFrom != nil {
			q = q.Filter(
				s.model.ProcessState.StatusChanged.Gte(filter.StatusChangedFrom),
			)
		}
		if filter.StatusChangedTo != nil {
			q = q.Filter(
				s.model.ProcessState.StatusChanged.Lte(filter.StatusChangedTo),
			)
		}
	}

	if len(filter.EnvelopeID) != 0 {

		q = q.Filter(qb.Or(
			qb.Exists(
				s.model.Route.GetMapper().Table().Select(qb.SQLText("1")).
					Where(s.model.Route.ProcessID.Eq(s.model.Process.ID),
						s.model.Route.EnvelopeID.In(
							uuidSliceToInterfaceSlice(filter.EnvelopeID)...),
					),
			),
			s.model.Process.TriggerEmitterID.In(
				uuidSliceToInterfaceSlice(filter.EnvelopeID)...),
		))
	}

	must(
		q.All(&dbProcessList))
	var pList []storage.Process
	for _, dbP := range dbProcessList {
		pList = append(pList, dbProcessToStorage(dbP))
	}
	return pList
}

// FindProcessForEnvelope finds the process(es) that reference a bunch of envelopes
func (s *ProcessStorage) FindProcessForEnvelope(
	envelopeIDs []api.UUID,
) map[api.UUID]api.UUID {
	result := make(map[api.UUID]api.UUID)
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.Route).
		Select(s.model.Route.EnvelopeID, s.model.Route.ProcessID).
		Where(s.model.Route.EnvelopeID.In(uuidSliceToInterfaceSlice(envelopeIDs)...))

	rows, err := q.SQLQuery()
	must(err)
	defer rows.Close()
	for rows.Next() {
		var eID, pID uuid.UUID
		must(rows.Scan(&eID, &pID))
		result[api.UUID(eID)] = api.UUID(pID)
	}

	// Some process may be route less, we need their trigger envelope ids
	q = tx.Query(s.model.Process).
		Select(s.model.Process.TriggerEnvelopeID, s.model.Process.ID).
		Where(s.model.Process.TriggerEnvelopeID.In(uuidSliceToInterfaceSlice(envelopeIDs)...))

	rows, err = q.SQLQuery()
	must(err)
	defer rows.Close()
	for rows.Next() {
		var eID, pID uuid.UUID
		must(rows.Scan(&eID, &pID))
		result[api.UUID(eID)] = api.UUID(pID)
	}

	return result
}

// CreateProcess saves a process
func (s *ProcessStorage) CreateProcess(p storage.Process) (api.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	p.ID = api.UUID{}

	dbProcess := storageProcessToDB(p)

	must(tx.Insert(&dbProcess))

	tx.MustCommit()

	return api.UUID(dbProcess.ID), nil
}

// SetGraph updates the graph & sourceOutput of a process
func (s *ProcessStorage) SetGraph(
	processID api.UUID,
	graphID api.UUID,
	sourceOutput storage.NodeOutputRef,
) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	p, ok := s.loadProcess(tx.Tx, processID)
	if !ok {
		panic("Unknown processID")
	}
	p.GraphID = graphID
	p.SourceOutput = sourceOutput
	dbP := storageProcessToDB(p)
	must(tx.Update(
		&dbP,
		model.ProcessGraphID,
		model.ProcessSourceNodeID,
		model.ProcessSourceSlotID))
	tx.MustCommit()
}

// GetTargets returns the targets for an envelope if already known, or 'nil' if not
func (s *ProcessStorage) GetTargets(
	processID api.UUID,
	source storage.NodeOutputRef,
	envelopeID api.UUID,
	noRouteTableUpdate bool,
	calcTargets func() ([]api.EnvelopeTarget, error),
) (targets []api.EnvelopeTarget, calcErr error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	bSource, err := json.Marshal(source)
	if err != nil {
		panic(err)
	}

	Route, RouteTarget := s.model.Route, s.model.RouteTarget

	var dbTargetList []model.RouteTarget

	if err := tx.Query(RouteTarget).
		LeftJoin(Route).
		Where(
			Route.ProcessID.Eq(uuid.UUID(processID)),
			Route.Source.Eq(bSource),
			Route.EnvelopeID.Eq(uuid.UUID(envelopeID))).
		All(&dbTargetList); err != nil {
		panic(err)
	}

	if len(dbTargetList) != 0 {
		for _, dbTarget := range dbTargetList {
			targets = append(targets, api.EnvelopeTarget{
				ActorID: api.UUID(dbTarget.ActorID),
				NodeID:  dbTarget.NodeID,
				Input:   dbTarget.Input,
			})
		}
		return
	}

	tx.Rollback()

	targets, calcErr = calcTargets()
	if calcErr != nil || noRouteTableUpdate {
		return
	}
	route := model.Route{
		ProcessID:  uuid.UUID(processID),
		Source:     string(bSource),
		EnvelopeID: uuid.UUID(envelopeID),
	}

	tx2 := s.MustBegin()
	defer tx2.RollbackIfOpened()

	// Use an exclusive transaction
	// on PG: SET TRANSACTION SERIALIZABLE
	if false && s.db.Engine.Dialect().Driver() == "postgres" {
		_, err := tx2.GetEngine().Exec(model.RawSQL("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE"))
		if err != nil {
			panic(err)
		}
	}

	if err := tx2.Query(RouteTarget).
		LeftJoin(Route).
		Where(
			Route.ProcessID.Eq(uuid.UUID(processID)),
			Route.Source.Eq(bSource),
			Route.EnvelopeID.Eq(uuid.UUID(envelopeID))).
		All(&dbTargetList); err != nil {
		panic(err)
	}

	// The routes where calculated faster by someone else
	// We return their result and ignore ours
	if len(dbTargetList) != 0 {
		for _, dbTarget := range dbTargetList {
			targets = append(targets, api.EnvelopeTarget{
				ActorID: api.UUID(dbTarget.ActorID),
				NodeID:  dbTarget.NodeID,
				Input:   dbTarget.Input,
			})
		}
		return
	}

	var logentries []storage.ProcessLogEntry
	// TODO: detect a concurrent insert and handle it gently
	must(tx2.Insert(&route))
	for _, target := range targets {
		must(tx2.Insert(&model.RouteTarget{
			RouteID: route.ID,
			ActorID: uuid.UUID(target.ActorID),
			NodeID:  target.NodeID,
			Input:   target.Input,
		}))
		logentry, err := s.processLogEmission(
			tx2,
			processID,
			storage.Emission{
				ProcessID:   processID,
				EnvelopeID:  envelopeID,
				TargetActor: target.ActorID,
				TargetInputRef: storage.NodeInputRef{
					NodeID: target.NodeID,
					Input:  target.Input,
				},
			},
			storage.EnvelopeStartPos())
		must(err)
		logentries = append(logentries, logentry)
	}
	s.logger.Noticef("New routes -> logentries = %+v", logentries)
	must(s.updateStateAndCommit(tx2, processID, logentries))
	return

}

// UpdateReceiveStatus updates a reception status
func (s *ProcessStorage) UpdateReceiveStatus(reception storage.Reception, status storage.ReceptionStatus) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogReception(tx, reception.ProcessID, reception, status)
	must(err)

	return s.updateStateAndCommit(
		tx, reception.ProcessID, []storage.ProcessLogEntry{logentry})
}

// ProcessNodeEnd signals the end of a ProcessNode.
func (s *ProcessStorage) ProcessNodeEnd(
	processID api.UUID, node storage.ProcessNode,
	envelopeID api.UUID, result api.ActorProcessingEndRequest_Result,
) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogProcessNodeEnd(
		tx, processID,
		storage.ProcessNodeEnd{
			Node:       node,
			EnvelopeID: envelopeID,
			Result:     result,
		})
	must(err)

	return s.updateStateAndCommit(
		tx, processID, []storage.ProcessLogEntry{logentry})
}

// GetEmitPosition returns an exclusive access to an emit position
func (s *ProcessStorage) GetEmitPosition(emission storage.Emission) (storage.EmitPosition, error) {
	var ownTx = true
	tx := s.MustBegin()
	defer func() {
		if ownTx {
			tx.Rollback()
		}
	}()
	// TODO Use an exclusive transaction
	// on PG: SET TRANSACTION SERIALIZABLE
	if false && s.db.Engine.Dialect().Driver() == "postgres" {
		_, err := tx.GetEngine().Exec(model.RawSQL("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE"))
		if err != nil {
			panic(err)
		}
	}
	// They load the value from the db
	var dbEmission model.Emission
	var pos = storage.EnvelopeStartPos()
	q := tx.Query(s.model.Emission).
		Where(
			s.model.Emission.ProcessID.Eq(uuid.UUID(emission.ProcessID)),
			s.model.Emission.EnvelopeID.Eq(uuid.UUID(emission.EnvelopeID)),
			s.model.Emission.TargetActor.Eq(uuid.UUID(emission.TargetActor)),
			s.model.Emission.TargetNodeID.Eq(emission.TargetInputRef.NodeID),
			s.model.Emission.TargetInput.Eq(emission.TargetInputRef.Input),
		).
		ForUpdate(s.model.Emission)
	if err := q.One(&dbEmission); err != nil {
		if err == yago.ErrRecordNotFound {
			// XXX The following code is pg specific
			_, insertErr := tx.GetEngine().Exec(model.RawSQL(
				"INSERT INTO emission "+
					"(process_id, envelope_id, target_actor, target_node_id, "+
					"target_input, start, complete, event_positions)"+
					" VALUES ($1, $2, $3, $4, $5, $6, $7, $8) "+
					" ON CONFLICT DO NOTHING",
				uuid.UUID(emission.ProcessID),
				uuid.UUID(emission.EnvelopeID),
				uuid.UUID(emission.TargetActor),
				emission.TargetInputRef.NodeID,
				emission.TargetInputRef.Input,
				true,
				false,
				"{}",
			))
			must(insertErr)
			// Retrieve the newly inserted emission or the one that was inserted
			// just before we did (in which case the insert should do nothing
			err = q.One(&dbEmission)
		}
		if err != nil {
			panic(err)
		}
	}
	pos.Start = dbEmission.Start
	pos.Complete = dbEmission.Complete
	if err := json.Unmarshal([]byte(dbEmission.EventPositions), &pos.Events); err != nil {
		panic(err)
	}
	ownTx = false
	return &EmitPosition{
		tx:             tx,
		emission:       dbEmission,
		pos:            pos,
		processStorage: s,
	}, nil
}

func (s *ProcessStorage) list(emissionClauses []qb.Clause, routeClauses []qb.Clause) []storage.Emission {
	Emission, Route, RouteTarget := s.model.Emission, s.model.Route, s.model.RouteTarget
	var dbEmissionList []model.Emission
	tx := s.MustBegin()
	defer tx.Rollback()
	must(
		tx.Query(Emission).
			Where(emissionClauses...).
			All(&dbEmissionList))
	var emissionList []storage.Emission
	for _, e := range dbEmissionList {
		emissionList = append(emissionList, storage.Emission{
			ProcessID:   api.UUID(e.ProcessID),
			EnvelopeID:  api.UUID(e.EnvelopeID),
			TargetActor: api.UUID(e.TargetActor),
			TargetInputRef: storage.NodeInputRef{
				NodeID: e.TargetNodeID,
				Input:  e.TargetInput,
			},
		})
	}

	q := tx.Query(RouteTarget).
		LeftJoin(Route).
		Where(routeClauses...).
		Filter(
			qb.NotExists(
				tx.Query(Emission).Where(
					Emission.ProcessID.Eq(Route.ProcessID),
					Emission.EnvelopeID.Eq(Route.EnvelopeID),
					Emission.TargetActor.Eq(RouteTarget.ActorID),
					Emission.TargetNodeID.Eq(RouteTarget.NodeID),
					Emission.TargetInput.Eq(RouteTarget.Input)).SelectStmt())).
		Select(
			Route.ProcessID,
			Route.EnvelopeID,
			RouteTarget.ActorID,
			RouteTarget.NodeID,
			RouteTarget.Input)

	stmt := q.SelectStmt().Build(s.db.Engine.Dialect())
	s.logger.Noticef("Listing unstarted routes: %s - %s",
		stmt.SQL(), stmt.Bindings(),
	)

	rows, err := q.SQLQuery()
	must(err)
	defer rows.Close()
	for rows.Next() {
		var emission storage.Emission
		must(rows.Scan((*uuid.UUID)(&emission.ProcessID),
			(*uuid.UUID)(&emission.EnvelopeID),
			(*uuid.UUID)(&emission.TargetActor),
			&emission.TargetInputRef.NodeID,
			&emission.TargetInputRef.Input))
		s.logger.Noticef("Adding dummy emission from unstarted route: %+v", emission)
		emissionList = append(emissionList, emission)
	}
	return emissionList
}

// ListActiveEmissions returns the active emissions
func (s *ProcessStorage) ListActiveEmissions(targetActor api.UUID) []storage.Emission {
	return s.list(
		[]qb.Clause{
			s.model.Emission.Complete.Eq(false),
			s.model.Emission.TargetActor.Eq(uuid.UUID(targetActor))},
		[]qb.Clause{
			s.model.RouteTarget.ActorID.Eq(uuid.UUID(targetActor)),
		},
	)
}

// ListProcessEmissions returns the active emissions
func (s *ProcessStorage) ListProcessEmissions(process api.UUID) []storage.Emission {
	return s.list(
		[]qb.Clause{
			s.model.Emission.Complete.Eq(false),
			s.model.Emission.ProcessID.Eq(uuid.UUID(process))},
		[]qb.Clause{
			s.model.Route.ProcessID.Eq(uuid.UUID(process)),
		},
	)
}

// GetState returns an instant copy of the state of a process
func (s *ProcessStorage) GetState(processID api.UUID) storage.ProcessState {
	tx := s.MustBegin()
	defer tx.Rollback()
	state, err := s.loadState(tx.Tx, processID, false)
	must(err)
	return state
}

// GetStateUpdater returns a StateUpdater for a process
func (s *ProcessStorage) GetStateUpdater(processID api.UUID) storage.StateUpdater {
	tx := s.MustBegin()
	updater, err := NewStateUpdater(s, tx, processID)
	if err != nil {
		panic(err)
	}
	return updater
}

// SetStatus changes the current status. Only Runnning->Paused and Paused->Running
// are allowed at this point
func (s *ProcessStorage) SetStatus(processID api.UUID, status api.Process_Status) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	process, ok := s.loadProcess(tx.Tx, processID)
	if !ok {
		return fmt.Errorf("No such process: %s", processID)
	}

	state, err := s.loadState(tx.Tx, processID, false)
	must(err)

	if !(status == api.Process_ERROR ||
		status == api.Process_PAUSED && state.Status == api.Process_RUNNING ||
		status == api.Process_RUNNING && state.Status == api.Process_PAUSED) {
		return fmt.Errorf("Invalid state transition: %s -> %s", state.Status, status)
	}

	old := state.Status
	state.Status = status

	err = s.saveState(tx.Tx, processID, state)
	must(err)

	tx.MustCommit()

	s.onStatusChanged(process, old, state.Status)

	return nil
}

func (s *ProcessStorage) loadState(tx *yago.Tx, processID api.UUID, forUpdate bool) (pState storage.ProcessState, err error) {
	var dbProcessState model.ProcessState

	q := tx.Query(s.model.ProcessState)
	if forUpdate {
		q = q.ForUpdate()
	}
	if err = q.Get(&dbProcessState, uuid.UUID(processID)); err != nil && err != yago.ErrRecordNotFound {
		return
	}

	if err == yago.ErrRecordNotFound {
		pState.Status = api.Process_INITIAL
	} else {
		pState.Status = dbProcessState.Status
		pState.StatusChanged = dbProcessState.StatusChanged
		pState.StatusReason = dbProcessState.StatusReason
		if err = json.Unmarshal(
			[]byte(dbProcessState.NodeStates),
			&pState.NodeStates); err != nil {
			return
		}
		pState.NodeStatesChanged = dbProcessState.NodeStatesChanged
	}
	return pState, nil
}

func (s *ProcessStorage) saveState(tx *yago.Tx, processID api.UUID, pState storage.ProcessState) error {
	b, err := json.Marshal(pState.NodeStates)
	if err != nil {
		return err
	}
	dbProcessState := model.ProcessState{
		ProcessID:         uuid.UUID(processID),
		Status:            pState.Status,
		StatusChanged:     pState.StatusChanged,
		StatusReason:      pState.StatusReason,
		NodeStates:        string(b),
		NodeStatesChanged: pState.NodeStatesChanged,
	}
	exists, err := tx.Query(s.model.ProcessState).
		Where(s.model.ProcessState.ProcessID.Eq(uuid.UUID(processID))).
		Exists()
	if err != nil {
		return err
	}
	if !exists {
		return tx.Insert(&dbProcessState)
	}
	return tx.Update(&dbProcessState)
}

func (s *ProcessStorage) processLogEmission(tx *TxHelper, processID api.UUID, emission storage.Emission, pos storage.EnvelopeReadPos) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		ProcessID:        processID,
		Emission:         &emission,
		EmissionPosition: &pos,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogReception(tx *TxHelper, processID api.UUID, reception storage.Reception, status storage.ReceptionStatus) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		ProcessID:       processID,
		Reception:       &reception,
		ReceptionStatus: status,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogProcessNodeEnd(
	tx *TxHelper, processID api.UUID, nodeEnd storage.ProcessNodeEnd,
) (storage.ProcessLogEntry, error) {
	return s.processLog(
		tx, processID,
		storage.ProcessLogEntry{
			ProcessID:      processID,
			ProcessNodeEnd: &nodeEnd,
		})
}

func (s *ProcessStorage) processLog(tx *TxHelper, processID api.UUID, logentry storage.ProcessLogEntry) (storage.ProcessLogEntry, error) {
	logentry.Timestamp = time.Now()
	b, err := json.Marshal(logentry)
	if err != nil {
		return storage.ProcessLogEntry{}, err
	}
	dbLogEntry := model.ProcessLogEntry{
		ProcessID: uuid.UUID(processID),
		Timestamp: logentry.Timestamp,
		Data:      string(b),
	}
	if err := tx.Insert(&dbLogEntry); err != nil {
		return storage.ProcessLogEntry{}, err
	}
	logentry.ID = strconv.FormatInt(dbLogEntry.ID, 10)
	return logentry, nil
}

func (s *ProcessStorage) updateStateAndCommit(tx *TxHelper, processID api.UUID, logentries []storage.ProcessLogEntry) error {
	// Commit the transaction so nextState is called outside it
	tx.MustCommit()

	s.processLogEntryHandler(logentries)
	return nil
}

// GetUnprocessedProcessLogEntries returns the unprocessed log entries
func (s *ProcessStorage) GetUnprocessedProcessLogEntries() (entries []storage.ProcessLogEntry) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbLogEntries []model.ProcessLogEntry

	q := tx.Query(s.model.ProcessLogEntry).
		Where(s.model.ProcessLogEntry.LogProcessed.Eq(false))
	must(q.All(&dbLogEntries))

	must(tx.Rollback())

	for _, dbLogEntry := range dbLogEntries {
		var logEntry storage.ProcessLogEntry
		if err := json.Unmarshal([]byte(dbLogEntry.Data), &logEntry); err != nil {
			panic(err)
		}
		logEntry.ID = strconv.FormatInt(dbLogEntry.ID, 10)
		logEntry.ProcessID = api.UUID(dbLogEntry.ProcessID)
		entries = append(entries, logEntry)
	}

	return
}

// QueryPostMortem returns a list of dead processes
func (s *ProcessStorage) QueryPostMortem(level api.LogLevel, statusList ...api.PMProcess_Status) []storage.Process {
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.Process).
		LeftJoin(s.model.ProcessState)
	if len(statusList) != 0 {
		var crits []qb.Clause
		for _, status := range statusList {
			crits = append(crits, s.model.ProcessState.PMStatus.Eq(status))
		}
		q = q.Filter(qb.Or(crits...))
	}
	if level == api.LogLevel_ERROR {
		q = q.Filter(s.model.ProcessState.PMLevel.Eq(api.LogLevel_ERROR))
	}

	var dbProcessList []model.Process
	must(
		q.All(&dbProcessList))
	var pList []storage.Process
	for _, dbP := range dbProcessList {
		pList = append(pList, dbProcessToStorage(dbP))
	}
	return pList
}

// GetPostMortemState returns the postmortem state of a process
func (s *ProcessStorage) GetPostMortemState(id api.UUID) storage.PostMortemState {
	tx := s.MustBegin()
	defer tx.Rollback()

	row := tx.Query(s.model.ProcessState).
		Where(s.model.ProcessState.ProcessID.Eq(uuid.UUID(id))).
		Select(
			s.model.ProcessState.PMLevel,
			s.model.ProcessState.PMStatus,
			s.model.ProcessState.PMComment,
			s.model.ProcessState.PMChanged).
		SQLQueryRow()

	var pmLevel, pmStatus []byte
	var pmState storage.PostMortemState

	must(row.Scan(&pmLevel, &pmStatus, &pmState.Comment, &pmState.Changed))
	must(pmState.Level.UnmarshalText(pmLevel))
	must(pmState.Status.UnmarshalText(pmStatus))

	return pmState
}

// SetPostMortemState sets the postmortem state of a process
func (s *ProcessStorage) SetPostMortemState(id api.UUID, pmState storage.PostMortemState) {
	tx := s.MustBegin()
	defer tx.MustCommit()

	pState := model.ProcessState{
		ProcessID: uuid.UUID(id),
		PMLevel:   pmState.Level,
		PMStatus:  pmState.Status,
		PMComment: pmState.Comment,
		PMChanged: pmState.Changed,
	}
	must(tx.Update(&pState,
		model.ProcessStatePMLevel,
		model.ProcessStatePMStatus,
		model.ProcessStatePMComment,
		model.ProcessStatePMChanged,
	))
}

// PurgeProcess removes all data of a process.
// The process must be terminated
func (s *ProcessStorage) PurgeProcess(id api.UUID) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	_, exists := s.loadProcess(tx.Tx, id)
	if !exists {
		return fmt.Errorf("No such process: %s", id)
	}
	state, err := s.loadState(tx.Tx, id, true)
	if state.Status != api.Process_DONE && state.Status != api.Process_ERROR {
		return fmt.Errorf("Process %s is not terminated (current status: %s). Cannot purge", id, state.Status)
	}

	// Drop process. The db FK will do the rest
	_, err = tx.GetEngine().Exec(
		s.model.Process.GetMapper().Table().Delete().
			Where(s.model.Process.ID.Eq(uuid.UUID(id))))
	must(err)

	tx.MustCommit()

	return nil
}

// EmitPosition implements storage.EmitPosition
type EmitPosition struct {
	tx             *TxHelper
	emission       model.Emission
	pos            storage.EnvelopeReadPos
	processStorage *ProcessStorage
	logentries     []storage.ProcessLogEntry
	lastError      error
}

// Get returns the position
func (p *EmitPosition) Get() (storage.EnvelopeReadPos, error) {
	return p.pos, p.lastError
}

// Set updates the position
func (p *EmitPosition) Set(pos storage.EnvelopeReadPos) error {
	if p.lastError != nil {
		return p.lastError
	}
	p.pos = pos
	p.emission.Start = pos.Start
	p.emission.Complete = pos.Complete
	b, err := json.Marshal(pos.Events)
	p.lastError = err
	must(err)
	p.emission.EventPositions = string(b)

	p.lastError = p.tx.Update(
		&p.emission,
		model.EmissionStart,
		model.EmissionComplete,
		model.EmissionEventPositions)
	must(p.lastError)
	logentry, err := p.processStorage.processLogEmission(
		p.tx,
		api.UUID(p.emission.ProcessID),
		storage.Emission{
			ProcessID:   api.UUID(p.emission.ProcessID),
			EnvelopeID:  api.UUID(p.emission.EnvelopeID),
			TargetActor: api.UUID(p.emission.TargetActor),
			TargetInputRef: storage.NodeInputRef{
				NodeID: p.emission.TargetNodeID,
				Input:  p.emission.TargetInput,
			},
		},
		pos)
	p.lastError = err
	must(err)
	p.logentries = append(p.logentries, logentry)
	return nil
}

// Unlock releases the lock on the position
func (p *EmitPosition) Unlock() error {
	if p.lastError != nil || len(p.logentries) == 0 {
		p.tx.Rollback()
		return nil
	}
	defer p.tx.RollbackIfOpened()
	return p.processStorage.updateStateAndCommit(
		p.tx, api.UUID(p.emission.ProcessID), p.logentries)
}

// StateUpdater implements storage.StateUpdater
type StateUpdater struct {
	processStorage *ProcessStorage
	tx             *TxHelper
	processID      api.UUID
	state          storage.ProcessState

	emission   model.Emission
	pos        storage.EnvelopeReadPos
	logentries []storage.ProcessLogEntry
	lastError  error
}

// NewStateUpdater creates a new StateUpdater
func NewStateUpdater(
	processStorage *ProcessStorage, tx *TxHelper, processID api.UUID) (*StateUpdater, error) {
	state, err := processStorage.loadState(tx.Tx, processID, true)
	if err != nil {
		return nil, err
	}
	return &StateUpdater{
		processStorage: processStorage,
		tx:             tx,
		processID:      processID,
		state:          state,
	}, nil
}

// Get returns the current state
func (u *StateUpdater) Get() (storage.ProcessState, error) {
	return u.state, nil
}

// Set updates the state
func (u *StateUpdater) Set(state storage.ProcessState) error {
	err := u.processStorage.saveState(u.tx.Tx, u.processID, state)
	if err != nil {
		return err
	}
	u.state = state
	return nil
}

// MarkLogEntry marks a ProcessLogEntry as processed
func (u *StateUpdater) MarkLogEntry(id string, withError bool) error {
	dbID, err := strconv.ParseInt(id, 10, 64)
	if err != nil {
		return err
	}
	return u.tx.Update(
		&model.ProcessLogEntry{ID: dbID, LogProcessed: true, ProcessingError: withError},
		model.ProcessLogEntryLogProcessed,
		model.ProcessLogEntryProcessingError,
	)
}

// Unlock release the lock on the data
func (u *StateUpdater) Unlock() error {
	return u.tx.Commit()
}
