// Copyright 2017 Orus.io. All rights reserved.

package xbus

import (
	"context"
	"errors"
	"fmt"
	"time"

	"github.com/nats-io/go-nats"

	api "bitbucket.org/orus-io/go-xbus/api"
)

// EnvelopeHandler is a type of callbacks for when new envelope arrive on an input
type EnvelopeHandler func(*EnvelopeReceiver)

// RawEnvelopeHandler is a type of callback for envelope handlers
type RawEnvelopeHandler func(input string, envelope api.Envelope) (api.EnvelopeAck, error)

// newActor initialize an actor
func newActor(client *Client, options ActorOptions) *Actor {
	actor := Actor{
		ID:      options.ID,
		Name:    options.Name,
		Kind:    options.Kind,
		Options: options,
		Client:  client,
	}
	actor.Log = &ActorLogger{
		NewPrefixedLogger("[actor/"+options.Name+"] ", client.Log),
		&actor}
	return &actor
}

// Actor provides the common tools for all actors
type Actor struct {
	ID      api.UUID
	Name    string
	Kind    api.Actor_Kind
	Options ActorOptions
	Log     *ActorLogger
	Client  *Client
}

// GetStringSetting returns a string typed setting entry
func (a *Actor) GetStringSetting(name string) string {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return ""
	}
	sValue, ok := value.(string)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value: %s", value))
	}
	return sValue
}

// MustGetStringSetting returns a string setting or panic if undefined
func (a *Actor) MustGetStringSetting(name string) string {
	value := a.GetStringSetting(name)
	if value == "" {
		a.Client.Log.Fatalf("Missing setting: %s", name)
	}
	return value
}

// GetIntSettingD returns a int setting, or the defaultValue if not specified
func (a *Actor) GetIntSettingD(name string, defaultValue int) int {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return defaultValue
	}
	iValue, ok := value.(int)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value '%s'. Expected an integer", value))
	}
	return iValue
}

// GetClient returns the underlying Client instance
func (a *Actor) GetClient() *Client {
	return a.Client
}

// ProcessingEnd calls the ActorProcessingEnd xbus api
func (a *Actor) ProcessingEnd(
	context api.EnvelopeContext,
	envelopeID api.UUID,
	result api.ActorProcessingEndRequest_Result,
	messages []api.LogMessage,
) error {
	_, err := a.GetClient().tryProcessingEnd(
		a.ID, context, envelopeID, result, messages, true)

	return err
}

// ProcessingSuccess signal the success of a processing
// for actors with no outputs of it not all outputs were
// hit or closed
func (a *Actor) ProcessingSuccess(context api.EnvelopeContext, envelopeID api.UUID) error {
	return a.ProcessingEnd(context, envelopeID, api.ActorProcessingEndRequest_SUCCESS, nil)
}

// ProcessingError signals a processing error.
func (a *Actor) ProcessingError(context api.EnvelopeContext, envelopeID api.UUID, format string, fmtArgs ...interface{}) error {
	return a.ProcessingEnd(context, envelopeID, api.ActorProcessingEndRequest_ERROR, []api.LogMessage{
		api.LogMessage{
			Time:  time.Now(),
			Level: api.LogLevel_ERROR,
			Text:  fmt.Sprintf(format, fmtArgs...),
		}})
}

// SendEnvelope send a potentially incomplete envelope to the broker
func (a *Actor) SendEnvelope(output string, env api.Envelope) (status api.EnvelopeAck_ReceptionStatus, err error) {
	if a.Kind == api.Actor_CONSUMER {
		return api.EnvelopeAck_ERROR, fmt.Errorf("A 'Consumer' actor cannot emit to outputs")
	}
	return a.Client.SendEnvelope(a.ID, output, env)
}

// SendItems send a group of items to the broker
// The envelope and event will be constructed automatically
func (a *Actor) SendItems(output, eventType string, items []api.Item) (status api.EnvelopeAck_ReceptionStatus, err error) {
	return a.SendEnvelope(output, ForgeEnvelopeFromItems(eventType, items))
}

// SendItemsContext send a group of items to the broker, in a given context
// The envelope and event will be constructed automatically
func (a *Actor) SendItemsContext(context api.EnvelopeContext, output, eventType string, items []api.Item) (status api.EnvelopeAck_ReceptionStatus, err error) {
	envelope := ForgeEnvelopeFromItems(eventType, items)
	envelope.Context = context
	return a.SendEnvelope(output, envelope)
}

// ForgeEnvelopeFromEvent ...
func ForgeEnvelopeFromEvent(event api.Event) api.Envelope {
	envID, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	return api.Envelope{
		ID:       envID,
		EventIDs: []api.UUID{event.ID},
		Events:   []api.Event{event},
	}
}

// ForgeEnvelopeFromItems ...
func ForgeEnvelopeFromItems(eventType string, items []api.Item) api.Envelope {
	evID, err := api.UUIDV4()
	if err != nil {
		panic(err)
	}
	checksum := api.Checksum(0).
		Update(evID[:16]).
		Update([]byte(eventType))

	for _, item := range items {
		checksum = checksum.Update(item)
	}

	event := api.Event{
		ID:        evID,
		Type:      eventType,
		Index:     1,
		Items:     items,
		ItemCount: uint64(len(items)),
		Checksum:  checksum,
	}
	return ForgeEnvelopeFromEvent(event)
}

// RawSubscription is what RawSubscribe returns
type RawSubscription interface {
	Unsubscribe() error
}

type rawActorInputServer struct {
	handler RawEnvelopeHandler
}

func (s rawActorInputServer) Input(
	ctx context.Context, input, envelopeID string, envelope api.Envelope,
) (api.EnvelopeAck, error) {
	return s.handler(input, envelope)
}

// RawSubscribe subscribes to the raw envelope stream
// The handler must:
// - validates the envelope integrity itself
// - send a EnvelopeAck to the 'reply' subject.
// If wildcard is false, the handler will have the responsibility to
// subscribe to the envelope dedicated subject to get all the remaining
// fragments.
// If wildcard is true, all the envelope subjects will be subscribed too,
// meaning the handler will receive not only the first but any fragment
// of the envelopes.
// In the case of multiple subscription on the same Actor (in case of
// redondant processes or multiple local handlers), using the wildcard
// implies a carreful handling of parallelism, as a single envelope fragments
// can be sent to different handlers.
func (a *Actor) RawSubscribe(handler RawEnvelopeHandler, wildcard bool) (RawSubscription, error) {
	if a.Kind == api.Actor_EMITTER {
		return nil, fmt.Errorf("A 'Emitter' actor cannot subscribe to inputs")
	}
	var subject string
	if wildcard {
		subject = api.ActorInputSubject(a.ID, ">")
	} else {
		subject = api.ActorInputSubject(a.ID, "*")
	}
	s := api.NewActorInputHandler(
		context.Background(),
		a.Client.conn,
		rawActorInputServer{handler})
	sub, err := a.Client.conn.QueueSubscribe(
		subject,
		a.ID.String(),
		s.Handler,
	)
	if err != nil {
		return nil, err
	}
	a.Client.API.Sentinel.ActorReady(a.ID.String())
	return sub, nil
}

// Subscribe setup the subscription
func (a *Actor) Subscribe(handler EnvelopeHandler) (*Subscription, error) {
	s := Subscription{
		actor:   a,
		handler: handler,
	}
	subject := api.ActorInputSubject(a.ID, "*.new")
	sub, err := a.GetClient().conn.QueueSubscribe(
		subject, a.ID.String(), api.NewActorInputHandler(
			context.Background(), a.GetClient().conn, &s,
		).Handler,
	)
	if err != nil {
		return nil, err
	}
	s.sub = sub
	a.Client.API.Sentinel.ActorReady(a.ID.String())
	return &s, nil
}

// OpenOutput returns a EnvelopeSender
func (a *Actor) OpenOutput(
	context api.EnvelopeContext, output string, eventTypes ...string,
) *EnvelopeSender {
	var es *EnvelopeSender
	es = NewEnvelopeSender(
		context, eventTypes, api.EnvelopeMaxSize,
		func(e api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			return a.SendEnvelope(output, e)
		})
	return es
}

// Subscription holds the nats subcription and call the handler when
// an envelope starts arriving.
// It implements the api.ActorInputServer interface
type Subscription struct {
	actor   *Actor
	handler EnvelopeHandler
	sub     *nats.Subscription
}

// Unsubscribe unsuscribes the nats handler
func (s *Subscription) Unsubscribe() error {
	return s.sub.Unsubscribe()
}

// Input implements the ActorInput.Input API
// It receives the first fragment of the envelopes sent to the actor,
// and when needed setup a receiver for the remaining fragments.
func (s *Subscription) Input(
	ctx context.Context, input string, envelopeSID string, env api.Envelope,
) (api.EnvelopeAck, error) {
	if envelopeSID != "new" {
		panic("This handler should receive only envelope heads")
	}

	s.actor.GetClient().Log.Debugf("Received new envelope head on input %s", input)

	er, err := NewEnvelopeReceiver(input, env)
	if err != nil {
		return *api.NewEnvelopeAckError(env.ID, err), nil
	}

	ack := api.EnvelopeAck{ID: env.ID, Status: er.Status()}

	if er.Status() == api.EnvelopeAck_RECEIVING {
		// Envelope was not complete, we need to listen for next messages
		// until envelope is complete or receiver stops receiving

		// TODO replace e.ID.String() with a function of envelope context + ID
		var tr = tailReceiver{
			er:     er,
			client: s.actor.GetClient(),
		}
		subject := api.ActorInputSubject(s.actor.ID, input) + "." + env.ID.String()

		// TODO Handle case were no more envelopes are received -> currently
		// leads to a leak and a never closed receiver.
		// For that, we will need to keep track of all receivers and their
		// subscriptions with a 'lastReceived' timestamp for periodic cleanup
		// This list of subs/receivers will also be useful with an actor is taken
		// down so we can signal the receiver that it's over (or reinitialize it
		// after a reconnection).
		sub, err := s.actor.Client.conn.Subscribe(
			subject,
			api.NewActorInputHandler(context.Background(), s.actor.Client.conn, &tr).Handler,
		)
		if err != nil {
			return api.EnvelopeAck{}, err
		}
		tr.sub = sub

	}
	go s.handler(er)
	return ack, nil
}

type tailReceiver struct {
	er     *EnvelopeReceiver
	sub    *nats.Subscription
	client *Client
}

func (r tailReceiver) Input(
	ctx context.Context, input string, envelopeSID string, env api.Envelope,
) (api.EnvelopeAck, error) {
	if envelopeSID != r.er.ID.String() {
		panic("This handler should receive the tail of the " + r.er.ID.String() + " envelope")
	}

	defer func() {
		if r.er.Status() != api.EnvelopeAck_RECEIVING {
			// Reception is over, we can unsubscribe
			if err := r.sub.Unsubscribe(); err != nil {
				r.client.Log.Errorf("Error unsubscribing: %s", err)
			}
		}
	}()

	switch status := r.er.Status(); status {
	case api.EnvelopeAck_RECEIVING:
		err := r.er.addEnvelope(env)

		if err != nil {
			return *api.NewEnvelopeAckError(env.ID, err), nil
		}
		return *api.NewEnvelopeAck(env.ID, r.er.Status()), nil
	case api.EnvelopeAck_ACCEPTED:
		return *api.NewEnvelopeAckError(
			env.ID, errors.New("Envelope already accepted"),
		), nil
	case api.EnvelopeAck_ERROR:
		ack := *api.NewEnvelopeAck(env.ID, api.EnvelopeAck_ERROR)
		ack.Reason = "Reception was already in error state"
		return ack, nil
	default:
		panic("Unknown reception status: " + status.String())
	}
}
