// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"
	"path"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus"
	api "bitbucket.org/orus-io/go-xbus/api"
)

// Logger is a logger automatically configured by command line options
var Logger xbus.Logger

var (
	cfgFile string

	// Viper is Viper instance that rootCmd (and children) will use
	Viper = viper.New()
)

// RootCmd represents the base command when called without any subcommands
var RootCmd = &cobra.Command{
	Use:   "xbus-client",
	Short: "Xbus client utility",
	Long: `Xbus client utility

xbus-client is a utility for managing a xbus client.`,
}

// Execute adds all child commands to the root command sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	if err := RootCmd.Execute(); err != nil {
		fmt.Println(err)
		os.Exit(-1)
	}
}

func init() {
	cobra.OnInitialize(initConfig)

	flags := RootCmd.PersistentFlags()

	flags.StringVar(&cfgFile, "config", "", "config file (default is ./xbus-client.yaml")
	flags.BoolP("debug", "d", false, "Enable debug mode")
	flags.Bool("trace", false, "Enable trace mode")

	Viper.BindPFlag("logging.debug", flags.Lookup("debug"))
	Viper.BindPFlag("logging.trace", flags.Lookup("trace"))

	xbus.DefineAndBindPFlags(Viper, flags)
}

// GetClient returns a Client based on the configuration file and flags
func GetClient() (*xbus.Client, error) {
	configDir, configName := ".", "xbus-client"
	if cfgFile != "" {
		configDir = path.Dir(cfgFile)
		configName = path.Base(cfgFile)
		ext := path.Ext(configName)
		if ext != ".yaml" {
			panic("The configuration file extension must be '.yaml'")
		}
		configName = configName[:len(configName)-len(ext)]
	}
	opts := xbus.COptionsFromViper(
		Viper, api.Account_ACTOR, Logger, configDir, configName, true)
	c := xbus.NewClient(opts, Logger)
	return c, nil
}

// initConfig reads in config file and ENV variables if set.
// it also setup the logger
func initConfig() {
	if cfgFile != "" { // enable ability to specify config file via flag
		Viper.SetConfigFile(cfgFile)
	} else {
		Viper.SetConfigName("xbus-client") // name of config file (without extension)
	}

	Viper.SetEnvPrefix("XBUSC")
	Viper.AddConfigPath(".")
	Viper.AutomaticEnv()

	Logger = xbus.GetLogger(&xbus.LoggerOptions{
		Debug: Viper.GetBool("logging.debug"),
		Trace: Viper.GetBool("logging.trace"),
	})

	// If a config file is found, read it in.
	err := Viper.ReadInConfig()
	if _, ok := err.(viper.ConfigFileNotFoundError); ok {
		Logger.Noticef(
			"Configuration file not found. May create one in './xbus-client.yaml'",
		)
	} else {
		switch {
		case os.IsNotExist(err):
			Logger.Noticef(
				"Configuration file not found. May create one in '%s'", cfgFile,
			)
		case err == nil:
			Logger.Debugf("Using config file: %s", Viper.ConfigFileUsed())
		default:
			Logger.Debugf("%#v", err)
			Logger.Fatalf("Could not read configuration file. Error was '%s'", err)
		}
	}
}
